//----------------------------------------------------------------------------------
// File:   Denoising.fx
// Author: Alexander Kharlamov
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

Texture2D txNoisy; 
Texture2D txClean; 
Texture2D txRestored;

float2    g_Shift;
float2    g_TexScale;
float2    g_TexShift;
float     g_NoiseLevel;
float     g_WeightThreshold;
float     g_LerpCoefficient;
float     g_CounterThreshold;
float     g_GaussianSigma;
float     g_Width;
float     g_Height;
float     g_WindowRadius;
float     g_BlockRadius;

SamplerState samLinear
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float4 Pos : POSITION;
    float2 Tex : TEXCOORD;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
};

//--------------------------------------------------------------------------------------
// Vertex Shaders
//--------------------------------------------------------------------------------------
// This Vertex shader is used to draw part of the picture
// Texture coordinates are scaled and shifted so that an image larger then 320x408 can be scrolled
PS_INPUT VS( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    output.Pos = input.Pos;
    output.Tex = g_TexShift + input.Tex * g_TexScale;
    return output;
}
// This Vertex shader is used to draw a fullscreen quad
PS_INPUT VSRestored( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    output.Pos = input.Pos;
    output.Tex = input.Tex;
    return output;
}
//--------------------------------------------------------------------------------------
// Pixel Shader Clean
//--------------------------------------------------------------------------------------
float4 PSClean( PS_INPUT input) : SV_Target
{
    float4 f4Tex00 = txClean.Sample( samLinear, input.Tex );
    return f4Tex00;
}
//--------------------------------------------------------------------------------------
// Pixel Shader Noisy
//--------------------------------------------------------------------------------------
float4 PSNoisy( PS_INPUT input) : SV_Target
{
    float4 f4Tex00 = txNoisy.Sample( samLinear, input.Tex );
    return f4Tex00;
}
//--------------------------------------------------------------------------------------
// Pixel Shader Restored
//--------------------------------------------------------------------------------------
float4 PSRestored( PS_INPUT input) : SV_Target
{
    float4 f4Tex00 = txRestored.Sample( samLinear, input.Tex );
    return f4Tex00;
}
//--------------------------------------------------------------------------------------
// Pixel Shader NLM_dynamic_loop
//--------------------------------------------------------------------------------------
float4 PSNLM_dynamic_loop( PS_INPUT input) : SV_Target
{
//  This is a straight forward implementation of denoising algorithm called "Non Local Means"
//  The idea behind this method is pretty simple: bluring removes white noise (to some extent) as it is a low pass filter
//  The restored pixel is computed as:
//  Rp(x,y) = Sum( P(x+i, y+j) * Weight(x, y, i, j)); i,j = [-WindowRadius, WindowRadius]
//  Weight in case of NLM is computed as
//  Weight(x, y, i, j) = exp( Sum( ColorDistance(P(x+i+n, y+j+m), P(x+n, y+m)) ) / g_NoiseLevel ); n,m = [-BlockRadius, BlockRadius]
//  ColorDistance is a function that should count the color distance between P1, P2 
//  Here a simple SE is used:
//  ColorDistance( P, Q ) = SQR(P.r - Q.r) + SQR(P.g - Q.g) + SQR(P.b - Q.b);
    float4 f4Result = {0.0f, 0.0f, 0.0f, 1.0f};
    float  fCounter = 0.0f;
    float  fWeight;  
    float  fSum = 0.0f;
    
    float i, j, n, m;
    
    float invBlockArea = 2.0f * g_BlockRadius + 1.0f;
    invBlockArea = 1.0f / (invBlockArea * invBlockArea);
    
    for (i = -g_WindowRadius; i <= g_WindowRadius; i++)
        for (j = -g_WindowRadius; j <= g_WindowRadius; j++)
        {                        
				fWeight = 0.0f;
                float3 f3TexIJc;
                float2 tex2 = input.Tex + float2(0, 0);
				for (n = -g_BlockRadius; n <= g_BlockRadius; n++)
                {
					for (m = -g_BlockRadius; m <= g_BlockRadius; m++)
					{
							float3 f3TexIJb = txNoisy.Sample( samLinear, tex2 + g_Shift * float2(i+n, j+m));                
							f3TexIJc.rgb    = txNoisy.Sample( samLinear, tex2 + g_Shift * float2(  n,   m));                    
							float3 f3Weight = f3TexIJb - f3TexIJc;
							f3Weight = f3Weight * f3Weight;
							fWeight = fWeight + f3Weight.r + f3Weight.g + f3Weight.b;
					}    
                }
				f3TexIJc = txNoisy.Sample( samLinear, input.Tex + g_Shift * float2(i, j));

				fWeight = fWeight * invBlockArea;
				// To ensure blurring in smooth areas, lets also add the usual gaussian blurring as well
				// NOTE: gaussian blurring is NOT actually needed. Changing g_NoiseLevel is enough to achieve good visual results
				fWeight = exp( - ( fWeight * g_NoiseLevel + (i * i + j * j) * g_GaussianSigma) );
	            
				fCounter += (fWeight > g_WeightThreshold) ? 1.0f: 0.0f;
	            
				fSum = fSum + fWeight;
	            
				f4Result.rgb = f4Result.rgb + f3TexIJc.rgb * fWeight;
        }            
//  That's how many pixels were avareged to compute restored pixel
    float iWindowArea = 2.0f * g_WindowRadius + 1.0f;
    iWindowArea = iWindowArea * iWindowArea;

    float3 f3Tex00 = txNoisy.Sample( samLinear, input.Tex ).rgb;

    f4Result.rgb = f4Result.rgb / fSum;	
//  Now the restored pixel is ready
//  But maybe the area is actually edgy and so it's better to take the pixel from the original image?	
//  This test shows if the area is smooth or not
	float lerpQ = (fCounter > (g_CounterThreshold * iWindowArea)) ? 1.0f-g_LerpCoefficient : g_LerpCoefficient;
//  This is the last lerp
//  Most common values for g_LerpCoefficient = [0.85, 1];
//  So if the area is smooth the result will be
//  RestoredPixel*0.85 + NoisyImage*0.15
//  If the area is noisy
//  RestoredPixel*0.15 + NoisyImage*0.85
//  That allows to preserve edges more thoroughly
    f4Result.rgb = lerp(f4Result.rgb, f3Tex00, lerpQ);
    
    return f4Result;        
}
//--------------------------------------------------------------------------------------
// Pixel Shader NLM_show_edges
//--------------------------------------------------------------------------------------
float4 PSNLM_show_edges( PS_INPUT input) : SV_Target
{
//  This pixel shader is an almost exact copy of PSNLM_dynamic_loop
//  However it only counts the number of pixels that "look like" the current pixel
//  If that number is high enough - that means the area around current pixel is smooth
//  Otherwise it is considered to be edgy and it is best to keep the noisy value
//  This pixel shader returns RED if the area is smooth and BLUE otherwise
//  It should help to choose 
//  - g_WeightThreshold
//  - g_CounterThreshold
	float  fCounter = 0.0f;
    float  fWeight;  
    
    float i, j, n, m;
    
    float invBlockArea = 2.0f * g_BlockRadius + 1.0f;
    invBlockArea = 1.0f / (invBlockArea * invBlockArea);
    
    for (i = -g_WindowRadius; i <= g_WindowRadius; i++)
        for (j = -g_WindowRadius; j <= g_WindowRadius; j++)
        {                        
				fWeight = 0.0f;
                float3 f3TexIJc;
                float2 tex2 = input.Tex + float2(0, 0);
				for (n = -g_BlockRadius; n <= g_BlockRadius; n++)
                {
					for (m = -g_BlockRadius; m <= g_BlockRadius; m++)
					{
							float3 f3TexIJb = txNoisy.Sample( samLinear, tex2 + g_Shift * float2(i+n, j+m));                
							f3TexIJc.rgb    = txNoisy.Sample( samLinear, tex2 + g_Shift * float2(  n,   m));                    
							float3 f3Weight = f3TexIJb - f3TexIJc;
							f3Weight = f3Weight * f3Weight;
							fWeight = fWeight + f3Weight.r + f3Weight.g + f3Weight.b;
					}    
                }
				fWeight = fWeight * invBlockArea;
				// To ensure blurring in smooth areas, lets also add the usual gaussian blurring as well
				// NOTE: gaussian blurring is NOT actually needed. Changing g_NoiseLevel is enough to achieve good visual results
				fWeight = exp( - ( fWeight * g_NoiseLevel + (i * i + j * j) * g_GaussianSigma) );
	            
				fCounter += (fWeight > g_WeightThreshold) ? 1.0f: 0.0f;	           
        }            
            
    float iWindowArea = 2.0f * g_WindowRadius + 1.0f;
    iWindowArea = iWindowArea * iWindowArea;

    float4 f4Result = {0.0f, 0.0f, 0.0f, 1.0f};
    // return RED if the area is smooth
    // return BLUE if the area is edgy
    f4Result.rgb = (fCounter > (g_CounterThreshold * iWindowArea) ) ? float3(1.0f, 0.0f, 0.0f) : float3(0.0f, 0.0f, 1.0f);
    
    return f4Result;        
}
//--------------------------------------------------------------------------------------
// Pixel Shader KNN_dynamic_loop
//--------------------------------------------------------------------------------------
float4 PSKNN_dynamic_loop( PS_INPUT input) : SV_Target
{
//  This is a straight forward implementation of denoising algorithm called "K Nearest Neighbors" a.k.a. Bilateral, SUSAN, etc.
//  The idea behind this method is pretty simple: bluring removes white noise (to some extent) as it is a low pass filter
//  The restored pixel (Rp) is computed as:
//  Rp(x,y) = Sum( P(x+i, y+j) * Weight(x, y, i, j)); i,j = [-WindowRadius, WindowRadius]
//  Weight in case of KNN is computed as
//  Weight(x, y, i, j) = ColorDistance(P(x+i, y+j), P(x, y));
//  ColorDistance is a function that should count the color distance between P1, P2 
//  Here a simple SE is used:
//  ColorDistance( P, Q ) = SQR(P.r - Q.r) + SQR(P.g - Q.g) + SQR(P.b - Q.b);
    float3 f3Tex00 = txNoisy.Sample( samLinear, input.Tex ).rgb;
    float3 f3TexIJ;
    float3 f3Weight;
    float4 f4Result = {0.0f, 0.0f, 0.0f, 1.0f};
    float  fCounter = 0.0f; 
    float  fWeight;        
    float  fSum = 0.0f;
    
    float i, j;
    float iWindowArea = 2.0f * g_WindowRadius + 1.0f;
    iWindowArea = iWindowArea * iWindowArea;
    
    for (i = -g_WindowRadius; i <= g_WindowRadius; i++)
        for (j = -g_WindowRadius; j <= g_WindowRadius; j++)
        {            
            f3TexIJ = txNoisy.Sample( samLinear, input.Tex + g_Shift * float2(i, j) ).rgb;            
            f3Weight = f3Tex00.rgb - f3TexIJ.rgb;
            f3Weight = f3Weight * f3Weight;
            fWeight = f3Weight.r + f3Weight.g + f3Weight.b;

			// To ensure blurring in smooth areas, lets also add the usual gaussian blurring as well
			// NOTE: gaussian blurring is NOT actually needed. Changing g_NoiseLevel is enough to achieve good visual results
            fWeight = exp( - ( fWeight * g_NoiseLevel + (i * i + j * j) * g_GaussianSigma ) );
            fCounter += (fWeight > g_WeightThreshold) ? 1.0f: 0.0f;
            
            fSum += fWeight;
            
            f4Result.rgb += f3TexIJ * fWeight;
        }            
            
    f4Result.rgb = f4Result.rgb / fSum;	
//  Now the restored pixel is ready
//  But maybe the area is actually edgy and so it's better to take the pixel from the original image?	
//  This test shows if the area is smooth or not
	float lerpQ = (fCounter > (g_CounterThreshold * iWindowArea)) ? 1.0f-g_LerpCoefficient : g_LerpCoefficient;
//  This is the last lerp
//  Most common values for g_LerpCoefficient = [0.85, 1];
//  So if the area is smooth the result will be
//  RestoredPixel*0.85 + NoisyImage*0.15
//  If the area is noisy
//  RestoredPixel*0.15 + NoisyImage*0.85
//  That allows to preserve edges more thoroughly
    f4Result.rgb = lerp(f4Result.rgb, f3Tex00, lerpQ);    
    return f4Result;        
}

//--------------------------------------------------------------------------------------
// Pixel Shader KNN_show_edges
//--------------------------------------------------------------------------------------
float4 PSKNN_show_edges( PS_INPUT input) : SV_Target
{
//  This pixel shader is an almost exact copy of PSNLM_dynamic_loop
//  However it only counts the number of pixels that "look like" the current pixel
//  If that number is high enough - that means the area around current pixel is smooth
//  Otherwise it is considered to be edgy and it is best to keep the noisy value
//  This pixel shader returns RED if the area is smooth and BLUE otherwise
//  It should help to choose 
//  - g_WeightThreshold
//  - g_CounterThreshold
    float3 f3Tex00 = txNoisy.Sample( samLinear, input.Tex ).rgb;
    float3 f3TexIJ;
    float3 f3Weight;
    float4 f4Result = {0.0f, 0.0f, 0.0f, 1.0f};
    float  fCounter = 0.0f; 
    float  fWeight;        
    
    float i, j;
    float iWindowArea = 2.0f * g_WindowRadius + 1.0f;
    iWindowArea = iWindowArea * iWindowArea;
    
    for (i = -g_WindowRadius; i <= g_WindowRadius; i++)
        for (j = -g_WindowRadius; j <= g_WindowRadius; j++)
        {            
            f3TexIJ = txNoisy.Sample( samLinear, input.Tex + g_Shift * float2(i, j) ).rgb;            
            f3Weight = f3Tex00.rgb - f3TexIJ.rgb;
            f3Weight = f3Weight * f3Weight;
            fWeight = f3Weight.r + f3Weight.g + f3Weight.b;

			// To ensure blurring in smooth areas, lets also add the usual gaussian blurring as well
			// NOTE: gaussian blurring is NOT actually needed. Changing g_NoiseLevel is enough to achieve good visual results
            fWeight = exp( - ( fWeight * g_NoiseLevel + (i * i + j * j) * g_GaussianSigma ) );
            fCounter += (fWeight > g_WeightThreshold) ? 1.0f: 0.0f;                    
        }            

    // return RED if the area is smooth
    // return BLUE if the area is edgy
    f4Result.rgb = (fCounter > (g_CounterThreshold * iWindowArea) ) ? float3(1.0f, 0.0f, 0.0f) : float3(0.0f, 0.0f, 1.0f);
    
    return f4Result;        
}

DepthStencilState DisableDepthTest
{
    DepthEnable = false;
    DepthWriteMask = 0;
};
//--------------------------------------------------------------------------------------
technique10 Simple
{
    pass Clean
    {
    
        SetDepthStencilState( DisableDepthTest, 0 ); // I dont need depth or stencil test for my algorithm
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSClean() ) );
    }
    pass Noisy
    {
        SetDepthStencilState( DisableDepthTest, 0 );   
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSNoisy() ) );
    }

    pass Restored
    {
    
        SetDepthStencilState( DisableDepthTest, 0 );
        SetVertexShader( CompileShader( vs_4_0, VSRestored() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSRestored() ) );
    }

}

//--------------------------------------------------------------------------------------
technique10 tKNN
{
    pass KNN
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSKNN_dynamic_loop() ) );
    }
    pass Area
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSKNN_show_edges() ) );
    }
    pass KNNRestored
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VSRestored() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSKNN_dynamic_loop() ) );
    }
    pass AreaRestored
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VSRestored() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSKNN_show_edges() ) );
    }
}
//--------------------------------------------------------------------------------------
technique10 tNLM
{
    pass NLM
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSNLM_dynamic_loop() ) );
    }
    pass Area
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSNLM_show_edges() ) );
    }
    pass NLMRestored
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VSRestored() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSNLM_dynamic_loop() ) );
    }
    pass AreaRestored
    {
        SetDepthStencilState( DisableDepthTest, 0 ); 
        SetVertexShader( CompileShader( vs_4_0, VSRestored() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSNLM_show_edges() ) );
    }
}